import zope.interface
import zope.component

import plone.registry.field

from zope.schema.interfaces import IField, IChoice
from zope.schema.interfaces import ISource, IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary

from plone.registry.interfaces import IPersistentField
from plone.registry.field import DisallowedProperty, StubbornProperty, InterfaceConstrainedProperty
from plone.registry.field import is_primitive


@zope.interface.implementer(IPersistentField)
@zope.component.adapter(IField)
def persistentFieldAdapter(context):
    """Turn a non-persistent field into a persistent one
    """
    
    if IPersistentField.providedBy(context):
        return context
    
    # See if we have an equivalently-named field
    
    class_name = context.__class__.__name__
    persistent_class = getattr(plone.registry.field, class_name, None)
    if persistent_class is None:
        return None
    
    if not issubclass(persistent_class, context.__class__):
        __traceback_info__ = "Can only clone a field of an equivalent type."
        return None
    
    ignored = list(DisallowedProperty.uses + StubbornProperty.uses)
    constrained = list(InterfaceConstrainedProperty.uses)
    
    instance = persistent_class.__new__(persistent_class)
    
    context_dict = dict([(k,v) for k,v in context.__dict__.items() 
                            if k not in ignored])

    for k,iface in constrained:
        v = context_dict.get(k, None)
        if v is not None and v != context.missing_value:
            v = iface(v, None)
            if v is None:
                __traceback_info__ = "The property `%s` cannot be adapted to `%s`." % (k, iface.__identifier__,)
                return None
            context_dict[k] = v

    instance.__dict__.update(context_dict)
    return instance

@zope.interface.implementer(IPersistentField)
@zope.component.adapter(IChoice)
def choicePersistentFieldAdapter(context):
    """Special handling for Choice fields.
    """
    
    instance = persistentFieldAdapter(context)
    if instance is None:
        return None
    
    if ISource.providedBy(context.vocabulary) or \
            IContextSourceBinder.providedBy(context.vocabulary):
        
        safe = False
        
        # Attempt to reverse engineer a 'values' argument
        if isinstance(context.vocabulary, SimpleVocabulary):
            values = []
            safe = True
            for term in context.vocabulary:
                if term.token == str(term.value) and is_primitive(term.value):
                    values.append(term.value)
                else:
                    safe = False
                    break
            if safe:
                instance._values = values
        
        if not safe:
            __traceback_info__ = "Persistent fields only support named vocabularies " + \
                                    "or vocabularies based on simple value sets."
            return None

    return instance


from persistent import Persistent

from zope.interface import implements, alsoProvides
from zope.event import notify

from zope.dottedname.resolve import resolve

from plone.registry.interfaces import IPersistentField
from plone.registry.interfaces import IRecord, IInterfaceAwareRecord

from plone.registry.events import RecordModifiedEvent

class FieldValidatedProperty(object):
    
    def __init__(self, name, initial_value):
        self._name = name
        self._initial_value = initial_value
        
    def __get__(self, inst, type_=None):
        return inst.__dict__.get(self._name, self._initial_value)
    
    def __set__(self, inst, value):
        if inst.field is None:
            raise ValueError("The record's field must be set before its value")
        
        field = inst.field.bind(inst)
        if value != field.missing_value:
            field.validate(value)
        
        oldValue = self.__get__(inst)
        inst.__dict__[self._name] = value
        
        notify(RecordModifiedEvent(inst, oldValue, value))

_marker = object()

class Record(Persistent):
    """A record that is stored in the registry.
    """
    
    implements(IRecord)

    __name__ = u""
    __parent__ = None

    field = None
    value = FieldValidatedProperty('value', None)

    def __init__(self, field, value=_marker, interface=None, fieldName=None):
        if not IPersistentField.providedBy(field):
            raise ValueError("Field is not persistent")

        # This lets field.set() work on the record
        field.__name__ = 'value'

        self.field = field
        
        if value is _marker:
            value = field.default
        else:
            bound_field = field.bind(self)
            if value != bound_field.missing_value:
                bound_field.validate(value)
        
        # Bypass event notification
        self.__dict__['value'] = value
        
        self.interfaceName = None
        if interface is not None:
            alsoProvides(self, IInterfaceAwareRecord)
            self.interfaceName = interface.__identifier__
        self.fieldName = fieldName
    
    @property
    def interface(self):
        try:
            return resolve(self.interfaceName)
        except ImportError:
            return None
    
    def __repr__(self):
        return "<Record %s>" % self.__name__


"""This module defines persistent versions of various fields.

The idea is that when a record is created, we copy relevant field properties
from the transient schema field from zope.schema, into the corresponding
persistent field. Note all field types are supported, but the common types
are.
"""

import persistent

import zope.interface

import zope.schema
import zope.schema._field
import zope.schema.vocabulary
import zope.schema.interfaces

_missing_value_marker = object()

from plone.registry.interfaces import IPersistentField

def is_primitive(value):
    return value is None or \
            isinstance(value, (int, long, bool, str, unicode, tuple, list,
                                set, frozenset, dict, float))

class DisallowedProperty(object):
    """A property that may not be set on an instance. It may still be set
    defined in a base class.
    """

    uses = []
    
    def __init__(self, name):
        self._name = name
        DisallowedProperty.uses.append(name)
    
    def __get__(self, inst, type_=None):
        # look for the object in bases
        if type_ is not None:
            for c in type_.__mro__:
                if self._name in c.__dict__ and not \
                        isinstance(c.__dict__[self._name], DisallowedProperty):
                    function = c.__dict__[self._name]
                    return function.__get__(inst, type_)
        raise AttributeError(self._name)
        
    def __set__(self, inst, value):
        raise ValueError(u"Persistent fields does not support setting the `%s` property" % self._name)

class StubbornProperty(object):
    """A property that stays stubbornly at a single, pre-defined value.
    """

    uses = []

    def __init__(self, name, value):
        StubbornProperty.uses.append(name)
        self._name = name
        self._value = value

    def __set__(self, inst, value):
        pass
        
    def __get__(self, inst, type_=None):
        return self._value

class InterfaceConstrainedProperty(object):
    """A property that may only contain values providing a certain interface.
    """
    
    uses = []
    
    def __init__(self, name, interface):
        InterfaceConstrainedProperty.uses.append((name, interface))
        self._name = name
        self._interface = interface
        
    def __set__(self, inst, value):
        if value != inst.missing_value:
            if not self._interface.providedBy(value):
                raise ValueError(u"The property `%s` may only contain objects providing `%s`." % 
                                    (self._name, self._interface.__identifier__,))
        inst.__dict__[self._name] = value

class PersistentField(persistent.Persistent):
    """Base class for persistent field definitions.
    """
    
    zope.interface.implements(IPersistentField)
    
    # Persistent fields do not have an order
    order = StubbornProperty('order', -1)
    
    # We don't allow setting a custom constraint, as this would introduce a
    # dependency on a symbol such as a function that may go away
    constraint = DisallowedProperty('constraint')

class PersistentCollectionField(PersistentField, zope.schema._field.AbstractCollection):
    """Ensure that value_type is a persistent field
    """
    
    value_type = InterfaceConstrainedProperty('value_type', IPersistentField)
    
class Bytes(PersistentField, zope.schema.Bytes):
    pass

class BytesLine(PersistentField, zope.schema.BytesLine):
    pass

class ASCII(PersistentField, zope.schema.ASCII):
    pass

class ASCIILine(PersistentField, zope.schema.ASCIILine):
    pass

class Text(PersistentField, zope.schema.Text):
    pass

class TextLine(PersistentField, zope.schema.TextLine):
    pass
    
class Bool(PersistentField, zope.schema.Bool):
    pass

class Int(PersistentField, zope.schema.Int):
    pass

class Float(PersistentField, zope.schema.Float):
    pass

class Tuple(PersistentCollectionField, zope.schema.Tuple):
    pass
    
class List(PersistentCollectionField, zope.schema.List):
    pass

class Set(PersistentCollectionField, zope.schema.Set):
    pass
    
class FrozenSet(PersistentCollectionField, zope.schema.FrozenSet):
    pass

class Password(PersistentField, zope.schema.Password):
    pass

class Dict(PersistentField, zope.schema.Dict):
    
    key_type = InterfaceConstrainedProperty('key_type', IPersistentField)
    value_type = InterfaceConstrainedProperty('value_type', IPersistentField)
    
class Datetime(PersistentField, zope.schema.Datetime):
    pass
    
class Date(PersistentField, zope.schema.Date):
    pass
    
class Timedelta(PersistentField, zope.schema.Timedelta):
    pass

class SourceText(PersistentField, zope.schema.SourceText):
    pass
    
class URI(PersistentField, zope.schema.URI):
    pass
    
class Id(PersistentField, zope.schema.Id):
    pass

class DottedName(PersistentField, zope.schema.DottedName):
    pass
    
class Choice(PersistentField, zope.schema.Choice):
    
    # We can only support string name or primitive=list vocabularies
    
    _values = None
    _vocabulary = None
    
    def __init__(self, values=None, vocabulary=None, source=None, **kw):
        
        if vocabulary is not None and not isinstance(vocabulary, basestring):
            raise ValueError("Persistent fields only support named vocabularies "
                             "or vocabularies based on simple value sets.")
        elif source is not None:
            raise ValueError("Persistent fields do not support sources, only named "
                             "vocabularies or vocabularies based on simple value sets.")

        assert not (values is None and vocabulary is None), (
               "You must specify either values or vocabulary.")
        assert values is None or vocabulary is None, (
               "You cannot specify both values and vocabulary.")

        self.vocabularyName = None
        
        if values is not None:
            
            for value in values:
                if not is_primitive(value):
                    raise ValueError("Vocabulary values may only contain primitive values.")
            
            self._values = values
        else:
            self.vocabularyName = vocabulary            
        
        self._init_field = bool(self.vocabularyName)
        super(zope.schema.Choice, self).__init__(**kw)
        self._init_field = False

    @property
    def vocabulary(self):
        # may be set by bind()
        if self._vocabulary is not None:
            return self._vocabulary
        if self._values is not None:
            return zope.schema.vocabulary.SimpleVocabulary.fromValues(self._values)
    DisallowedProperty.uses.append('vocabulary')
    
    # override bind to allow us to keep constraints on the 'vocabulary'
    # property
    def bind(self, object):
        clone = zope.schema.Field.bind(self, object)
        # get registered vocabulary if needed:
        if zope.schema.interfaces.IContextSourceBinder.providedBy(self.vocabulary):
            clone._vocabulary = self.vocabulary(object)
            assert zope.schema.interfaces.ISource.providedBy(clone.vocabulary)
        elif clone.vocabulary is None and self.vocabularyName is not None:
            vr = zope.schema.vocabulary.getVocabularyRegistry()
            clone._vocabulary = vr.get(object, self.vocabularyName)
            assert zope.schema.interfaces.ISource.providedBy(clone.vocabulary)
        return clone


import unittest
import doctest

from zope.testing import doctestunit
from zope.component import provideAdapter, testing, eventtesting

from plone.registry.fieldfactory import persistentFieldAdapter
from plone.registry.fieldfactory import choicePersistentFieldAdapter

from zope.interface import Interface
from zope import schema

class IMailSettings(Interface):
    """Settings for email
    """
    
    sender = schema.TextLine(title=u"Mail sender", default=u"root@localhost")
    smtp_host = schema.URI(title=u"SMTP host server")

class IMailPreferences(Interface):
    """Settings for email
    """
    
    max_daily = schema.Int(title=u"Maximum number of emails per day", min=0, default=3)
    settings = schema.Object(title=u"Mail setings to use", schema=IMailSettings)

def setUp(test=None):
    testing.setUp()
    eventtesting.setUp()
    
    provideAdapter(persistentFieldAdapter)
    provideAdapter(choicePersistentFieldAdapter)

class TestBugs(unittest.TestCase):
    """Regression tests for bugs that have been fixed
    """
    
    def setUp(self):
        setUp(self)
        
    def tearDown(self):
        testing.tearDown(self)
    
    def test_bind_choice(self):
        from plone.registry.field import Choice
        
        from zope.schema.vocabulary import getVocabularyRegistry
        from zope.schema.vocabulary import SimpleVocabulary
        
        def vocabFactory(obj):
            return SimpleVocabulary.fromValues(['one', 'two'])
        
        reg = getVocabularyRegistry()
        reg.register('my.vocab', vocabFactory)
        
        class T(object):
            f = None
        
        f = Choice(__name__='f', title=u"Test", vocabulary="my.vocab")
        t = T()
        
        # Bug: this would give "AttributeError: can't set attribute" on
        # clone.vocabulary
        f.bind(t)

        

def test_suite():
    return unittest.TestSuite([
        doctestunit.DocFileSuite(
            'registry.txt', package='plone.registry',
            optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            setUp=setUp, tearDown=testing.tearDown),
        doctestunit.DocFileSuite(
            'events.txt', package='plone.registry',
            optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            setUp=setUp, tearDown=testing.tearDown),
        doctestunit.DocFileSuite(
            'field.txt', package='plone.registry',
            optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            setUp=setUp, tearDown=testing.tearDown),
        unittest.makeSuite(TestBugs),
        ])



from zope.interface import implements, alsoProvides
from plone.registry.interfaces import IRecordsProxy

_marker = object()

class RecordsProxy(object):
    """A proxy that maps an interface to a number of records
    """
    
    implements(IRecordsProxy)
    
    def __init__(self, registry, schema, omitted=()):
        
        # skip __setattr__
        self.__dict__['__schema__'] = schema
        self.__dict__['__registry__'] = registry
        self.__dict__['__omitted__'] = omitted
        self.__dict__['__prefix__'] = schema.__identifier__ + '.'
        
        alsoProvides(self, schema)
        
    def __getattr__(self, name):
        if name not in self.__schema__:
            raise AttributeError(name)
        value = self.__registry__.get(self.__prefix__ + name, _marker)
        if value is _marker:
            value = self.__schema__[name].missing_value
        return value
        
    def __setattr__(self, name, value):
        if name in self.__schema__:
            full_name = self.__prefix__ + name
            if full_name not in self.__registry__:
                raise AttributeError(name)
            self.__registry__[full_name] = value
        else:
            self.__dict__[name] = value
    
    def __repr__(self):
        return "<RecordsProxy for %s>" % self.__schema__.__identifier__


from persistent import Persistent
from BTrees.OOBTree import OOBTree

from zope.interface import implements
from zope.component import queryAdapter
from zope.event import notify

from zope.schema import getFieldNames, getFieldsInOrder
from zope.schema.interfaces import InvalidDottedName
from zope.schema._field import _isdotted

from plone.registry.interfaces import IRegistry, IRecord, IPersistentField
from plone.registry.record import Record
from plone.registry.recordsproxy import RecordsProxy
from plone.registry.events import RecordAddedEvent, RecordRemovedEvent

class Registry(Persistent):
    """The persistent registry
    """
    
    implements(IRegistry)
    
    def __init__(self):
        self._records = Records(self)
    
    # Basic value access API
    
    def __getitem__(self, name):
        return self.records[name].value
    
    def get(self, name, default=None):
        record = self.records.get(name, None)
        if record is None:
            return default
        return record.value
    
    def __setitem__(self, name, value):
        self.records[name].value = value
    
    def __contains__(self, name):
        return name in self.records
    
    # Records - make this a property so that it's readonly
    
    @property
    def records(self):
        return self._records
    
    # Schema interface API
    
    def forInterface(self, interface, check=True, omit=()):
        prefix = interface.__identifier__ + '.'
        if check:
            for name in getFieldNames(interface):
                if name not in omit and prefix + name not in self.records:
                    raise KeyError("Interface `%s` defines a field `%s`, "
                                   "for which there is no record." % (interface.__identifier__, name))
        
        return RecordsProxy(self, interface, omitted=omit)

    def registerInterface(self, interface, omit=()):
        prefix = interface.__identifier__ + '.'
        for name, field in getFieldsInOrder(interface):
            if name in omit or field.readonly:
                continue
            record_name = prefix + name
            persistent_field = queryAdapter(field, IPersistentField)
            if persistent_field is None:
                raise TypeError("There is no persistent field equivalent for "
                                "the field `%s` of type `%s`." % (name, field.__class__.__name__))
            
            value = persistent_field.default
            
            # Attempt to retain the exisiting value
            if record_name in self.records:
                existing_record = self.records[record_name]
                value = existing_record.value
                bound_field = persistent_field.bind(existing_record)
                try:
                    bound_field.validate(value)
                except:
                    value = persistent_field.default
            
            self.records[record_name] = Record(persistent_field, value, 
                                               interface=interface, fieldName=name)

class Records(Persistent):
    """The records stored in the registry
    """
    
    __parent__ = None
    
    def __init__(self, parent):
        self.__parent__ = parent
        self.data = OOBTree()

    def __setitem__(self, name, record):
        if not _isdotted(name):
            raise InvalidDottedName(record)
        if not IRecord.providedBy(record):
            raise ValueError("Value must be a record")
        if not IPersistentField.providedBy(record.field):
            raise ValueError("The record's field must be an IPersistentField.")
        
        record.__name__ = name
        record.__parent__ = self.__parent__
        self.data[name] = record
        notify(RecordAddedEvent(record))
    
    def __delitem__(self, name):
        record = self[name]
        del self.data[name]
        notify(RecordRemovedEvent(record))

    # Unfortunately, you can't just subclass an OOBTree if you want to
    # persist it...

    def __getitem__(self, name):
        return self.data.__getitem__(name)
    
    def get(self, name, default=None):
        return self.data.get(name, default)
    
    def __nonzero__(self):
        return self.data.__nonzero__()
    
    def __len__(self):
        return self.data.__len__()
        
    def __iter__(self):
        return self.data.__iter__()
    
    def __getslice__(self, index1, index2):
        return self.data.__getslice__(index1, index2)
        
    def has_key(self, name):
        return self.data.has_key(name)
        
    def __contains__(self, name):
        return self.data.__contains__(name)
        
    def keys(self, min=None, max=None, excludemin=False, excludemax=False):
        return self.data.keys(min, max, excludemin, excludemax)
        
    def maxKey(self, key=None):
        return self.data.maxKey(key)
        
    def minKey(self, key=None):
        return self.data.minKey(key)
        
    def values(self, min=None, max=None, excludemin=False, excludemax=False):
        return self.data.values(min, max, excludemin, excludemax)
        
    def items(self, min=None, max=None, excludemin=False, excludemax=False):
        return self.data.items(min, max, excludemin, excludemax)
    
    def update(self, collection):
        self.data.update(collection)
        
    def byValue(self, minValue):
        self.data.byValue(minValue)
        
    def setdefault(self, key, d):
        return self.data.setdefault(key, d)
        
    def pop(self, key, d):
        return self.data.pop(key, d)
        
    def insert(self, key, value):
        self.data.insert(key, value)
        
    def difference(self, c1, c2):
        return self.data.difference(c1, c2)
        
    def union(self, c1, c2):
        return self.data.union(c1, c2)
    
    def intersection(self, c1, c2):
        return self.data.intersection(c1, c2)
    
    def clear(self):
        self.data.clear()


from setuptools import setup, find_packages
import os

version = '1.0b1'

setup(name='plone.registry',
      version=version,
      description="A debconf-like (or about:config-like) registry for storing application settings",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read() +
                       open(os.path.join("plone", "registry", "registry.txt")).read() +
                       open(os.path.join("plone", "registry", "events.txt")).read(),
                       # we skip field.txt, as it's very low level
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='configuration registry',
      author='Martin Aspeli, Wichert Akkerman, Hanno Schlichting',
      author_email='plone-developers@lists.sourceforge.net',
      url='http://pypi.python.org/pypi/plone.registry',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'ZODB3',
          'zope.schema',
          'zope.interface',
          'zope.component',
          'zope.event',
          'zope.testing',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



from zope.interface import implements
from zope.component import subscribers, adapter

from plone.registry.interfaces import IRecordEvent
from plone.registry.interfaces import IRecordAddedEvent
from plone.registry.interfaces import IRecordRemovedEvent
from plone.registry.interfaces import IRecordModifiedEvent
from plone.registry.interfaces import IInterfaceAwareRecord

from plone.registry.recordsproxy import RecordsProxy

class RecordEvent(object):
    implements(IRecordEvent)

    def __init__(self, record):
        self.record = record

    def __repr__(self):
        return "<%s for %s>" % (self.__class__.__name__, self.record.__name__)

class RecordAddedEvent(RecordEvent):
    implements(IRecordAddedEvent)

class RecordRemovedEvent(RecordEvent):
    implements(IRecordRemovedEvent)

class RecordModifiedEvent(RecordEvent):
    implements(IRecordModifiedEvent)
    
    def __init__(self, record, oldValue, newValue):
        super(RecordModifiedEvent, self).__init__(record)
        self.oldValue = oldValue
        self.newValue = newValue

@adapter(IRecordEvent)
def redispatchInterfaceAwareRecordEvents(event):
    """When an interface-aware record received a record event,
    redispatch the event in a simlar manner to the IObjectEvent redispatcher.
    
    Note that this means one IRecordModifiedEvent will be fired for each
    change to a record.
    """
    
    record = event.record
    
    if not IInterfaceAwareRecord.providedBy(record):
        return
    
    schema = record.interface
    if schema is None:
        return
    
    proxy = RecordsProxy(record.__parent__, schema)
    
    adapters = subscribers((proxy, event), None)
    for adapter in adapters:
        pass # getting them does the work


from zope.interface import Interface
from zope.interface.interfaces import IInterface

from zope import schema

from zope.schema.interfaces import IField

class IPersistentField(IField):
    """A field that can be persistent along with a record.
    
    We provide our own implementation of the basic field types that are
    supported by the registry.
    """
    
class IRecord(Interface):
    """A record stored in the registry.
    """
    
    field = schema.Object(title=u"A field describing this record",
                          schema=IPersistentField)
    
    value = schema.Field(title=u"The value of this record",
                         description=u"Must be valid according to the record's field")

class IRecordEvent(Interface):
    """Base interface for record level events
    """
    
    record = schema.Object(title=u"The record that was added.",
                           description=u"Both __name__ and __parent__ will be set before the event is fired",
                           schema=IRecord)

class IRecordAddedEvent(IRecordEvent):
    """Event fired when a record is added to a registry.
    """
    
class IRecordRemovedEvent(IRecordEvent):
    """Event fired when a record is removed from a registry.
    """
    
class IRecordModifiedEvent(IRecordEvent):
    """Event fired when a record's value is modified.
    """

    oldValue = schema.Field(title=u"The record's previous value")
    newValue = schema.Field(title=u"The record's new value")

class IInterfaceAwareRecord(Interface):
    """A record will be marked with this interface if it knows which
    interface its field came from.
    """
    
    interfaceName = schema.DottedName(title=u"Dotted name to interface")
    
    interface = schema.Object(title=u"Interface that provided the record",
                              description=u"May be None if the interface is no longer available",
                              schema=IInterface,
                              readonly=True)
    
    fieldName = schema.ASCIILine(title=u"Name of the field in the original interface")

class IRegistry(Interface):
    """The configuration registry
    """
    
    def __getitem__(key):
        """Get the value under the given key. A record must have been
        installed for this key for this to be valid. Otherwise, a KeyError is
        raised.
        """

    def get(key, default=None):
        """Attempt to get the value under the given key. If it does not
        exist, return the given default.
        """

        
    def __setitem__(key, value):
        """Set the value under the given key. A record must have been
        installed for this key for this to be valid. Otherwise, a KeyError is
        raised. If value is not of a type that's allowed by the record, a
        ValidationError is raised.
        """
        
    def __contains__(key):
        """Determine if the registry contains a record for the given key.
        """
        
    records = schema.Dict(
            title=u"The records of the registry",
            key_type=schema.DottedName(
                    title=u"Name of the record",
                    description=u"By convention, this should include the "
                                 "package name and optionally an interface "
                                 "named, if the record can be described by a "
                                 "field in an interface (see also "
                                 "registerInterface() below), e.g. "
                                 "my.package.interfaces.IMySettings.somefield.",
                ),
            value_type=schema.Object(
                    title=u"The record for this name",
                    schema=IRecord,
                ),
        )
        
    def forInterface(interface, check=True, omit=()):
        """Get an IRecordsProxy for the given interface. If `check` is True,
        an error will be raised if one or more fields in the interface does
        not have an equivalent setting.
        """

    def registerInterface(interface, omit=()):
        """Create a set of records based on the given interface. For each
        schema field in the interface, a record will be inserted with a
        name like `${interface.__identifier__}.${field.__name__}`, and a
        value equal to default value of that field. Any field with a name
        listed in `omit`, or with the `readonly` property set to True, will
        be ignored.
        """

class IRecordsProxy(Interface):
    """This object is returned by IRegistry.by_interface(). It will be
    made to provide the relevant interface, i.e. it will have the
    attributes that the interface promises. Those attributes will be retrieved
    from or written to the underlying IRegistry.
    """

    __schema__ = schema.Object(title=u"Interface providing records",
                               schema=IInterface,
                               readonly=True)

    __registry__ = schema.Object(title=u"Registry where records will be looked up",
                                 schema=IRegistry,
                                 readonly=True)

    __omitted__ = schema.Tuple(title=u"Fields that are not stored in the registry",
                               description=u"If any of these are accessed, you will get an AttributeError",
                               value_type=schema.ASCIILine(title=u"Fieldname"),
                               readonly=True)